#
# Copyright (C) 2020-2025 Soleta Networks <info@soleta.eu>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

import fdisk
from src.log import OgError

GPT_TO_GUID = {
    'LINUX-SWAP':   '0657FD6D-A4AB-43C4-84E5-0933C84B4F4F',
    'LINUX':        '0FC63DAF-8483-4772-8E79-3D69D8477DE4',
    'NTFS':         'EBD0A0A2-B9E5-4433-87C0-68B6B72699C7',
    'EFI':          'C12A7328-F81F-11D2-BA4B-00A0C93EC93B',
    'HFS':          '48465300-0000-11AA-AA11-00306543ECAC',
    'FAT32':        'EBD0A0A2-B9E5-4433-87C0-68B6B72699C7',
    'WIN-RECOV':    'DE94BBA4-06D1-4D40-A16A-BFD50179D6AC',
}

DOS_TO_TYPE = {
    'LINUX-SWAP':   0x82,
    'EXTENDED':     0x0f,
    'EMPTY':        0x00,
    'LINUX':        0x83,
    'CACHE':        0x83,
    'NTFS':         0x07,
    'HFS':          0xaf,
    'FAT32':        0x0b,
    'EFI':          0xef,
}

# mapping between GUID and internal database partition type
GUID_TO_TYPE = {
    # Windows
    'EBD0A0A2-B9E5-4433-87C0-68B6B72699C7': 0x0700,  # Microsoft basic data (NTFS)
    'E3C9E316-0B5C-4DB8-817D-F92DF00215AE': 0x0C01,  # Microsoft reserved
    'DE94BBA4-06D1-4D40-A16A-BFD50179D6AC': 0x2700,  # Windows recovery

    # EFI and boot related
    'C12A7328-F81F-11D2-BA4B-00A0C93EC93B': 0xEF00,  # EFI System Partition
    '21686148-6449-6E6F-744E-656564454649': 0xEF02,  # BIOS boot partition (grub)

    # Linux
    '0657FD6D-A4AB-43C4-84E5-0933C84B4F4F': 0x8200,  # Linux swap
    '0FC63DAF-8483-4772-8E79-3D69D8477DE4': 0x8300,  # Linux filesystem
    'E6D6D379-F507-44C2-A23C-238F2A3DF928': 0x8E00,  # Linux LVM

    # Apple
    '426F6F74-0000-11AA-AA11-00306543ECAC': 0xAB00,  # Recovery HD
    '48465300-0000-11AA-AA11-00306543ECAC': 0xAF00,  # HFS/HFS+
}

def get_dos_parttype(cxt, ptype_str):
    l = cxt.label
    code = DOS_TO_TYPE.get(ptype_str, 0x0)
    parttype = l.get_parttype_from_code(code)
    return parttype


def get_gpt_parttype(cxt, ptype_str):
    l = cxt.label
    uuid = GPT_TO_GUID.get(ptype_str, GPT_TO_GUID['LINUX'])
    parttype = l.get_parttype_from_string(uuid)
    return parttype


def get_parttype(cxt, ptype_str):
    if not cxt:
        raise OgError('No libfdisk context')
    if not cxt.label or cxt.label.name not in ['dos', 'gpt']:
        raise OgError('Unknown libfdisk label')
    if type(ptype_str) != str:
        raise OgError('Invalid partition type')

    if cxt.label.name == 'dos':
        return get_dos_parttype(cxt, ptype_str)
    elif cxt.label.name == 'gpt':
        return get_gpt_parttype(cxt, ptype_str)
    else:
        raise OgError(f'Invalid partition label \'{cxt.label.name}\'')
