#
# Copyright (C) 2020-2025 Soleta Networks <info@soleta.eu>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

from src.utils.winreg import *
from src.log import OgError
from src.utils.fs import *
import logging
import codecs
import shutil
import hivex
import os


POSTCONF_PATH = '/opt/opengnsys/postconf'
POSTCONF_MAIN_SCRIPT = 'postconf_master.cmd'

def check_postconf_scripts(script_list):
    missing_scripts = []
    for script_name in script_list:
        script_path = os.path.join(POSTCONF_PATH, script_name)

        if not os.path.isfile(script_path):
            missing_scripts.append(script_name)

    if missing_scripts:
        raise OgError(f'Cannot find post-configuration scripts: {", ".join(missing_scripts)}')

def _create_postconf_main_script(mountpoint, script_list):
    logging.info(f'Generating {POSTCONF_MAIN_SCRIPT}')
    main_file_path = os.path.join(mountpoint, 'Windows', 'System32', POSTCONF_MAIN_SCRIPT)

    if not os.path.isdir(os.path.dirname(main_file_path)):
        raise OgError(f'Target directory does not exist: {dir_path}')

    try:
        with open(main_file_path, 'w', encoding='utf-8', newline='\r\n') as f:
            f.write('@echo off\n')
            f.write('chcp 65001 >nul\n')
            f.write('set LOGFILE=C:\\postconf.log\n')
            f.write('echo ==== POSTCONF START ==== >> %LOGFILE%\n')
            f.write('echo Date: %DATE% Time: %TIME% >> %LOGFILE%\n')

            for script_name in script_list:
                f.write(f'echo Running {script_name} >> %LOGFILE%\n')
                abs_path = f'C:\\Windows\\System32\\{script_name}'
                if script_name.lower().endswith('.ps1'):
                    f.write(f'powershell.exe -ExecutionPolicy Bypass -File "{abs_path}" >> %LOGFILE% 2>>&1\n')
                else:
                    f.write(f'call "{abs_path}" >> %LOGFILE% 2>>&1\n')

            f.write('echo Finalizing setup >> %LOGFILE%\n')
            f.write('echo Disabling SystemSetupInProgress >> %LOGFILE%\n')
            f.write('REG ADD HKLM\\System\\Setup /v SystemSetupInProgress /t REG_DWORD /d 0 /f >> %LOGFILE% 2>>&1\n')
            f.write('echo Disabling CmdLine >> %LOGFILE%\n')
            f.write('REG ADD HKLM\\System\\Setup /v CmdLine /t REG_SZ /d "" /f >> %LOGFILE% 2>>&1\n')
            f.write('echo Disabling SetupType >> %LOGFILE%\n')
            f.write('REG ADD HKLM\\System\\Setup /v SetupType /t REG_DWORD /d 0 /f >> %LOGFILE% 2>>&1\n')
            f.write('echo ==== POSTCONF END ==== >> %LOGFILE%\n')
            f.write('echo Rebooting system >> %LOGFILE%\n')
            f.write('shutdown /r /t 30\n')
            f.write('exit /b 0\n')
    except OSError as e:
        raise OgError(f'Failed to create {POSTCONF_MAIN_SCRIPT}: {e}')


def _configure_system_postconf(mountpoint):
    logging.info('Configuring Windows mini-setup mode')
    system_hive_path = mountpoint + WINDOWS_HIVE_SYSTEM
    hive = hive_handler_open(system_hive_path, write=True)
    root = hive.root()

    setup_node = get_node_child(hive, root, 'Setup')

    hive.node_set_value(setup_node, {
        'key': 'SystemSetupInProgress',
        't': RegistryType.DWORD.value,
        'value': int(1).to_bytes(4, 'little'),
    })

    hive.node_set_value(setup_node, {
        'key': 'SetupType',
        't': RegistryType.DWORD.value,
        'value': int(4).to_bytes(4, 'little'),
    })

    cmd_line_str = f'cmd.exe /c C:\\Windows\\System32\\{POSTCONF_MAIN_SCRIPT}'
    hive.node_set_value(setup_node, {
        'key': 'CmdLine',
        't': RegistryType.SZ.value,
        'value': cmd_line_str.encode(WINDOWS_HIVE_ENCODING),
    })

    try:
        hive.commit(None)
    except Exception as e:
        raise OgError(f'Failed to commit changes to SYSTEM hive: {e}')

def _copy_postconf_scripts(mountpoint, script_list):
    logging.info(f'Installing post-configuration scripts: {", ".join(script_list)}')
    for script_name in script_list:
        postconf_file_src = os.path.join(POSTCONF_PATH, script_name)
        postconf_file_dst = os.path.join(mountpoint, 'Windows', 'System32', script_name)

        if not os.path.isfile(postconf_file_src):
            raise OgError(f'Postconf file {script_name} does not exist')

        try:
            shutil.copy(postconf_file_src, postconf_file_dst)
        except (OSError, shutil.SameFileError) as e:
            raise OgError(f'Failed to copy postconf script {script_name}: {e}')

def install_postconf_scripts(disk, partition, script_list):
    if not script_list:
        logging.info('No post-configuration scripts associated to this image')
        return

    logging.info('Post-configuration scripts detected...')
    check_postconf_scripts(script_list)

    device = get_partition_device(disk, partition)
    mountpoint = device.replace('dev', 'mnt')
    if not mount_mkdir(device, mountpoint):
        raise OgError(f'Unable to mount {device} into {mountpoint}')

    try:
        _copy_postconf_scripts(mountpoint, script_list)
        _create_postconf_main_script(mountpoint, script_list)
        _configure_system_postconf(mountpoint)
    finally:
        umount(mountpoint)
    logging.info('Post-configuration script will be run in the next system boot')
