#
# Copyright (C) 2020-2025 Soleta Networks <info@soleta.eu>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

import os
import logging
import shlex
import subprocess
import json
from uuid import UUID
from src.log import OgError

import fdisk


class ogPartitionData:
    def __init__(self, parttype, fstype, padev, size, partno):
        self.parttype = parttype
        self.fstype = fstype
        self.padev = padev
        self.size = size
        self.partno = partno

class ogDiskData:
    def __init__(self, nsectors, sector_size, label_name):
        self.nsectors = nsectors
        self.sector_size = sector_size
        self.label_name = label_name

def get_partition_data(device):
    res = []
    try:
        cxt = fdisk.Context(device=device, details=True, readonly=True)
    except Exception as e:
        raise OgError(f'Partition query error for {device}: {e}') from e

    for i, p in enumerate(cxt.partitions):
        pd = ogPartitionData(
            parttype = cxt.partition_to_string(p, fdisk.FDISK_FIELD_TYPEID),
            fstype = cxt.partition_to_string(p, fdisk.FDISK_FIELD_FSTYPE),
            padev = cxt.partition_to_string(p, fdisk.FDISK_FIELD_DEVICE),
            size = cxt.partition_to_string(p, fdisk.FDISK_FIELD_SIZE),
            partno = p.partno)
        res.append(pd)
    return res

def get_disk_data(device):
    try:
        cxt = fdisk.Context(device=device, details=True, readonly=True)
    except Exception as e:
        raise OgError(f'Disk query error for {device}: {e}') from e
    return ogDiskData(
            nsectors = cxt.nsectors,
            sector_size = cxt.sector_size,
            label_name = cxt.label.name if cxt.label else "")


def is_rotational(device_name):
    path = f'/sys/block/{device_name}/queue/rotational'
    try:
        with open(path, 'r') as f:
            return f.read().strip() == "1"
    except OSError as e:
        logging.error(f'Cannot read rotational value for device {device_name}: {e}')
        return False


def is_partition(device_name):
    if device_name.startswith(('sd', 'vd')):
        return device_name[-1].isdigit()
    if device_name.startswith('nvme'):
        parts = device_name.rsplit('p', 1)
        return len(parts) == 2 and parts[1].isdigit()
    return False


def get_disks():
    basedir_disk_by_path = '/dev/disk/by-path/'
    device_prefixes = ('sd', 'nvme', 'vd')
    device_map = {}

    try:
        device_list = os.listdir(basedir_disk_by_path)
    except FileNotFoundError:
        logging.error(f'Directory {basedir_disk_by_path} not found.')
        return []

    if not device_list:
        logging.error('No devices found in /dev/disk/by-path/')
        return []

    for symlink_name in device_list:
        symlink_path = os.path.join(basedir_disk_by_path, symlink_name)

        if not os.path.islink(symlink_path):
            continue

        device_path = os.readlink(symlink_path)
        device_name = os.path.basename(device_path)

        if not device_name.startswith(device_prefixes) or is_partition(device_name):
            continue

        # Store the shortest available symlink for each device.
        if device_name not in device_map or symlink_name < device_map[device_name][0]:
            device_map[device_name] = (symlink_name, device_name)

    disks = list(device_map.values())
    disks.sort(key=lambda x: (
        not x[1].startswith("nvme"),
        'usb' in x[0],
        is_rotational(x[1]),
        x[0])
    )

    return [device for _, device in disks]


def get_partition_device(disknum, partnum):
    """
    Returns the device path, given a disk and partition number
    """
    disk_list = get_disks()
    disk_index = disknum - 1
    if disk_index < 0 or disk_index >= len(disk_list):
        raise OgError(f'Invalid disk number {disknum}, {len(disk_list)} disks available.')

    disk = disk_list[disk_index]
    cxt = fdisk.Context(f'/dev/{disk}', readonly=True)

    for pa in cxt.partitions:
        if pa.partno == partnum - 1:
            return cxt.partition_to_string(pa, fdisk.FDISK_FIELD_DEVICE)

    raise OgError(f'No such partition with disk index {disknum} and partition index {partnum}')


def get_efi_partition(disknum, enforce_gpt):
    """
    Look for an EFI System Partition at the n-th disk. If disknum is invalid an
    exception is thrown.
    If enforce_gpt is set to True the ESP will be ignored in a MBR partition
    scheme.

    Returns tuple with:
        - Device name containing the ESP
        - /dev/{device} string
        - Partition number (starting at 1)
    """
    disk_index = disknum - 1
    if disk_index < 0 or disk_index >= len(get_disks()):
        raise OgError(f'Invalid disk number {disknum} when trying to find EFI partition, {len(get_disks())} disks available.')

    disk = get_disks()[disk_index]
    cxt = fdisk.Context(f'/dev/{disk}', readonly=True)

    if enforce_gpt and cxt.label.name == 'dos':
        raise OgError(f'Windows EFI partition requires GPT partition scheme, but /dev/{disk} has DOS partition scheme')

    logging.info('Searching EFI partition...')
    for pa in cxt.partitions:
        if pa.type.name == 'EFI System':
            logging.info(f'EFI partition found at /dev/{disk}')
            return cxt.partition_to_string(pa, fdisk.FDISK_FIELD_DEVICE), f'/dev/{disk}', pa.partno + 1
    raise OgError(f'Cannot find EFI partition at /dev/{disk}')


def get_partition_id(disk_index, part_index):
    device = get_partition_device(disk_index, part_index)
    cmd = f'blkid -s PARTUUID -o value {device}'
    proc = subprocess.run(shlex.split(cmd),
                          stdout=subprocess.PIPE,
                          encoding='utf-8')
    if proc.returncode != 0:
            raise OgError(f'failed to query partition UUID for {device}')
    return proc.stdout.strip()


def get_disk_id(disk_index):
    disk = get_disks()[disk_index - 1]
    disk_path = f'/dev/{disk}'
    cmd = f'blkid -s PTUUID -o value {disk_path}'
    proc = subprocess.run(shlex.split(cmd),
                          stdout=subprocess.PIPE,
                          encoding='utf-8')
    if proc.returncode != 0:
            raise OgError(f'failed to query disk UUID for {disk_path}')
    return proc.stdout.strip()


def get_filesystem_id(disk_index, part_index):
    device = get_partition_device(disk_index, part_index)
    cmd = f'blkid -s UUID -o value {device}'
    proc = subprocess.run(shlex.split(cmd),
                          stdout=subprocess.PIPE,
                          encoding='utf-8')
    if proc.returncode != 0:
            raise OgError(f'failed to query filesystem UUID for {device}')
    return proc.stdout.strip()


def get_sector_size(disk):
    disk_index = disk - 1

    if disk_index < 0 or disk_index >= len(get_disks()):
        raise OgError(f'Invalid disk number {disk} when trying to find ESP, {len(get_disks())} disks available.')

    device_name = get_disks()[disk_index]
    file_path = f'/sys/class/block/{device_name}/queue/hw_sector_size'

    try:
        with open(file_path, 'r') as f:
            data = f.read().strip()
    except OSError as e:
        raise OgError(f'Error while trying to read {file_path}: {e}') from e
    return int(data)


def get_partition_start_offset(disk, partition):
    disk_name = get_disks()[disk - 1]
    disk_path = f'/dev/{disk_name}'
    part_number = partition - 1

    cmd = f'sfdisk -J {disk_path}'
    proc = subprocess.run(shlex.split(cmd), capture_output=True, text=True)

    if proc.returncode != 0:
        raise OgError(f'Failed to query sfdisk')

    try:
        part_data_json = json.loads(proc.stdout)
    except json.JSONDecodeError as e:
        raise OgError(f'Invalid sfdisk output: {e}') from e

    try:
        part_data = part_data_json['partitiontable']['partitions']
        start_offset = part_data[part_number]['start']
    except KeyError as e:
        raise OgError(f'Error while trying to parse sfdisk: {e}') from e

    return start_offset


def uuid_to_bytes(uuid):
    uuid = uuid.replace('-', '')

    group0 = f'{uuid[6:8]}{uuid[4:6]}{uuid[2:4]}{uuid[0:2]}'
    group1 = f'{uuid[10:12]}{uuid[8:10]}'
    group2 = f'{uuid[14:16]}{uuid[12:14]}'
    group3 = uuid[16:20]
    group4 = uuid[20:32]
    res = f'{group0}-{group1}-{group2}-{group3}-{group4}'
    return UUID(res).bytes


def get_disk_id_bytes(disk):
    from src.utils.uefi import is_uefi_supported
    disk_id = get_disk_id(disk)

    if is_uefi_supported():
        return uuid_to_bytes(disk_id)

    return bytes.fromhex(disk_id)[::-1]


def get_part_id_bytes(disk, partition):
    from src.utils.uefi import is_uefi_supported
    if is_uefi_supported():
        part_id = get_partition_id(disk, partition)
        return uuid_to_bytes(part_id)

    partition_start_offset = get_partition_start_offset(disk, partition)
    sector_size = get_sector_size(disk)
    byte_offset = partition_start_offset * sector_size
    byte_offset = "{0:016x}".format(byte_offset)
    return bytes.fromhex(byte_offset)[::-1]
