#
# Copyright (C) 2020-2025 Soleta Networks <info@soleta.eu>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

import os
import shlex
import logging
import tempfile
import subprocess
from src.log import OgError


TMP_FILENAME = 'og_temp_script'

def get_interpreter_data():
    data = {
        'bash': {
            'template': 'bash "{SCRIPT_PATH}" {CMD_ARGS}',
            'extension': '.sh',
            'encoding': 'utf-8'
        },
        'cmd': {
            'template': 'cmd.exe /c "{SCRIPT_PATH}" {CMD_ARGS}',
            'extension': '.cmd',
            'encoding': 'utf-8'
        },
        'powershell': {
            'template': 'powershell.exe -ExecutionPolicy Bypass -Command \'[Console]::OutputEncoding = [System.Text.UTF8Encoding]::new(); & "{SCRIPT_PATH}" {CMD_ARGS}\'',
            'extension': '.ps1',
            'encoding': 'utf-8-sig'
        }
    }
    return data


def run_script(ogRest, cmd_args, script_content, interpreter):
    interpreter_data = get_interpreter_data()[interpreter]
    file_extension = interpreter_data['extension']
    script_encoding = interpreter_data['encoding']
    script_path = os.path.join(tempfile.gettempdir(), TMP_FILENAME + file_extension)

    try:
        with open(script_path, 'w', encoding=script_encoding) as f:
            f.write(script_content)
    except OSError as e:
        raise OgError(f'Error during script generation: {e}')

    cmd_template = interpreter_data['template']
    cmd = cmd_template.format(SCRIPT_PATH=script_path, CMD_ARGS=cmd_args)

    env = os.environ.copy()
    env['OutputEncoding'] = 'utf8'

    try:
        ogRest.proc = subprocess.run(
            shlex.split(cmd),
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            env=env,
        )
        output = ogRest.proc.stdout
    except (OSError, subprocess.SubprocessError) as e:
        raise OgError(f'Error when running command subprocess: {e}') from e

    output = output.decode('utf-8', errors='replace')
    return (ogRest.proc.returncode, output)